import pytest

from bonsai.active_directory import UserAccountControl


def test_init():
    """ Test creating UserAccountControl. """
    with pytest.raises(TypeError):
        _ = UserAccountControl("123")
    uac = UserAccountControl(209)
    assert uac is not None


def test_value():
    """ Test value property. """
    uac = UserAccountControl(209)
    assert uac.value == 209


def test_properties():
    """ Test properties property. """
    uac = UserAccountControl(514)
    assert uac.properties["accountdisable"]
    assert uac.properties["normal_account"]
    assert all(
        val == False
        for key, val in uac.properties.items()
        if key not in ("accountdisable", "normal_account")
    )
    uac = UserAccountControl(67584)
    assert uac.properties["dont_expire_password"]
    assert uac.properties["interdomain_trust_account"]
    assert all(
        val == False
        for key, val in uac.properties.items()
        if key not in ("dont_expire_password", "interdomain_trust_account")
    )



import math
import trio

from ..ldapconnection import BaseLDAPConnection, LDAPSearchScope
from ..errors import NotAllowedOnNonleaf, TimeoutError


class TrioLDAPConnection(BaseLDAPConnection):
    """
    Asynchronous LDAP connection object that works with trio.
    It has the same methods and properties as :class:`bonsai.LDAPConnection`, but
    with the exception of :meth:`bonsai.LDAPConnection.close` and
    :meth:`bonsai.LDAPConnection.fileno` all of them are awaitable.

    :param LDAPClient client: a client object.
    """

    def __init__(self, client):
        self.__open_coro = None
        super().__init__(client, is_async=True)

    async def __aenter__(self):
        """Async context manager entry point."""
        return await self.__open_coro

    async def __aexit__(self, type, value, traceback):
        """Async context manager exit point."""
        self.close()

    def __await__(self):
        return self.__open_coro.__await__()  # Hack to avoid returning a coroutine.

    def open(self, timeout=None):
        self.__open_coro = super().open(timeout)
        return self

    async def _poll(self, msg_id, timeout=None):
        tout_sec = timeout if timeout is not None else math.inf
        with trio.move_on_after(tout_sec):
            while True:
                await trio.lowlevel.wait_writable(self)
                await trio.lowlevel.wait_readable(self)
                res = super().get_result(msg_id)
                if res is not None:
                    return res
        raise TimeoutError("Timeout is exceeded")

    def _evaluate(self, msg_id, timeout=None):
        return self._poll(msg_id, timeout)

    async def delete(self, dname, timeout=None, recursive=False):
        try:
            return await super().delete(dname, timeout, recursive)
        except NotAllowedOnNonleaf as exc:
            if recursive:
                results = await self.search(
                    dname, LDAPSearchScope.ONELEVEL, attrlist=["1.1"], timeout=timeout
                )
                for res in results:
                    await self.delete(res.dn, timeout, True)
                return await self.delete(dname, timeout, False)
            else:
                raise exc

    async def _search_iter_anext(self, search_iter):
        try:
            return next(search_iter)
        except StopIteration:
            msgid = search_iter.acquire_next_page()
            if msgid is None:
                raise StopAsyncIteration from None
            search_iter = await self._evaluate(msgid)
            return next(search_iter)

    async def get_result(self, msg_id, timeout=None):
        return await self._evaluate(msg_id, timeout)



import struct

from typing import Dict, Optional

from .sid import SID
from .acl import ACL, ACE, ACEFlag, ACERight, ACEType, ACLRevision


class SecurityDescriptor:
    """
    A class that defines the security attributes of an object. These
    attributes specify who owns the object; who can access the object and
    what they can do with it; what level of audit logging can be applied to
    the object; and what kind of restrictions apply to the use of the
    security descriptor. It can contain two types of :class:`ACL`:

    * A discretionary access control list (DACL) is controlled by the
      owner of an object or anyone granted WRITE_DAC access to the object.
      It specifies the access particular users and groups can have to an
      object. For example, the owner of a file can use a DACL to control
      which users and groups can and cannot have access to the file.
    * A system access control list (SACL) is similar to the DACL, except
      that the SACL is used to audit rather than control access to an
      object. When an audited action occurs, the operating system records
      the event in the security log. Each ACE in a SACL has a header that
      indicates whether auditing is triggered by success, failure, or
      both; a SID that specifies a particular user or security group to
      monitor; and an access mask that lists the operations to audit.

    :param dict control: a dict that specifies control access bit flags.
    :param SID|None owner_sid: the SID of the owner of the object.
    :param SID|None group_sid: the SID of the group of the object.
    :param ACL|None sacl: the system access control list.
    :param ACL|None dacl: the discretionary access control list.
    :param int revision: the revision of the security descriptor.
    :param int sbz1: reserved value.
    """

    def __init__(
        self,
        control: Dict[str, bool],
        owner_sid: Optional[SID],
        group_sid: Optional[SID],
        sacl: Optional[ACL],
        dacl: Optional[ACL],
        revision: int = 1,
        sbz1: int = 0,
    ) -> None:
        self.__revision = revision
        self.__sbz1 = sbz1
        self.__control = control
        self.__owner_sid = owner_sid
        self.__group_sid = group_sid
        self.__sacl = sacl
        self.__dacl = dacl

    @staticmethod
    def __convert_ctrl(ctrl, to_int=False):
        values = {
            "self_relative": 0x8000,
            "rm_control_valid": 0x4000,
            "sacl_protected": 0x2000,
            "dacl_protected": 0x1000,
            "sacl_auto_inherited": 0x800,
            "dacl_auto_inherited": 0x400,
            "sacl_computed_inheritance_required": 0x200,
            "dacl_computed_inheritance_required": 0x100,
            "server_security": 0x80,
            "dacl_trusted": 0x40,
            "sacl_defaulted": 0x20,
            "sacl_present": 0x10,
            "dacl_defaulted": 0x8,
            "dacl_present": 0x4,
            "group_defaulted": 0x2,
            "owner_defaulted": 0x1,
        }
        if to_int:
            return sum(values[key] for key, val in ctrl.items() if val)
        else:
            return {key: bool(ctrl & val) for key, val in values.items()}

    @classmethod
    def from_binary(cls, data: bytes) -> "SecurityDescriptor":
        """
        Create a SecurityDescriptor object from a binary blob.

        :param bytes data: a little-endian byte ordered byte input.
        :returns: A new SecurityDescriptor instance.
        :rtype: SecurityDescriptor
        :raises TypeError: when the parameter is not bytes.
        :raises ValueError: when the input cannot be parsed as a 
            SecurityDescriptor object.
        """
        try:
            if not isinstance(data, bytes):
                raise TypeError("The `data` parameter must be bytes")
            sacl = None
            dacl = None
            (
                rev,
                sbz1,
                ctrl,
                offset_owner,
                offset_group,
                offset_sacl,
                offset_dacl,
            ) = struct.unpack("<BBHIIII", data[:20])
            ctrl = cls.__convert_ctrl(ctrl)
            owner_sid = SID(bytes_le=data[offset_owner:]) if offset_owner else None
            group_sid = SID(bytes_le=data[offset_group:]) if offset_group else None
            if ctrl["sacl_present"] and offset_sacl != 0:
                sacl = ACL.from_binary(data[offset_sacl:])
            if ctrl["dacl_present"] and offset_dacl != 0:
                dacl = ACL.from_binary(data[offset_dacl:])
            return cls(ctrl, owner_sid, group_sid, sacl, dacl, rev, sbz1)
        except struct.error as err:
            raise ValueError(f"Not a valid binary SecurityDescriptor, {err}")

    def to_binary(self) -> bytes:
        """
        Convert SecurityDescriptor object to binary form with little-endian byte order.

        :returns: Bytes of the binary SecurityDescriptor instance
        :rtype: bytes
        """
        owner = b""
        group = b""
        sacl = b""
        dacl = b""
        offset_owner = 0
        offset_group = 0
        offset_sacl = 0
        offset_dacl = 0
        ctrl = self.__convert_ctrl(self.control, True)
        data = bytearray(20)
        if self.owner_sid:
            owner = self.owner_sid.bytes_le
            offset_owner = 20
        if self.group_sid:
            group = self.group_sid.bytes_le
            offset_group = 20 + len(owner)
        if self.sacl:
            sacl = self.sacl.to_binary()
            offset_sacl = 20 + len(owner) + len(group)
        if self.dacl:
            dacl = self.dacl.to_binary()
            offset_dacl = 20 + len(owner) + len(group) + len(sacl)
        struct.pack_into(
            "<BBHIIII",
            data,
            0,
            self.revision,
            self.sbz1,
            ctrl,
            offset_owner,
            offset_group,
            offset_sacl,
            offset_dacl,
        )
        data.extend(owner)
        data.extend(group)
        data.extend(sacl)
        data.extend(dacl)
        return bytes(data)

    @property
    def sbz1(self) -> int:
        """ Reserved field in the security descriptor structure. """
        return self.__sbz1

    @property
    def revision(self) -> int:
        """ The revision of the security descriptor. """
        return self.__revision

    @property
    def control(self) -> Dict[str, bool]:
        """ The dict of the control access bit flags. """
        return self.__control

    @property
    def owner_sid(self) -> Optional[SID]:
        """ The :class:`SID` of the owner. """
        return self.__owner_sid

    @property
    def group_sid(self) -> Optional[SID]:
        """ The :class:`SID` of the group. """
        return self.__group_sid

    @property
    def sacl(self) -> Optional[ACL]:
        """ The system :class:`ACL`. """
        return self.__sacl

    @property
    def dacl(self) -> Optional[ACL]:
        """ The discretionary :class:`ACL`. """
        return self.__dacl


class UserAccountControl:
    """
    A class for parsing UserAccountControl field.

    :param int flags: integer representing the property flags.
    :raises TypeError: if flags parameter is not an int.
    """

    def __init__(self, flags: int) -> None:
        if not isinstance(flags, int):
            raise TypeError("The `flags` parameter must be an integer")
        self.__flag_values = {
            "script": 0x1,
            "accountdisable": 0x2,
            "homedir_required": 0x8,
            "lockout": 0x10,
            "passwd_notreqd": 0x20,
            "passwd_cant_change": 0x40,
            "encrypted_text_pwd_allowed": 0x80,
            "temp_duplicate_account": 0x100,
            "normal_account": 0x200,
            "interdomain_trust_account": 0x800,
            "workstation_trust_account": 0x1000,
            "server_trust_account": 0x2000,
            "dont_expire_password": 0x10000,
            "mns_logon_account": 0x20000,
            "smartcard_required": 0x40000,
            "trusted_for_delegation": 0x80000,
            "not_delegated": 0x100000,
            "use_des_key_only": 0x200000,
            "dont_req_preauth": 0x400000,
            "password_expired": 0x800000,
            "trusted_to_auth_for_delegation": 0x1000000,
            "partial_secrets_account": 0x4000000,
        }
        self.__properties = {
            key: bool(flags & val) for key, val in self.__flag_values.items()
        }

    @property
    def properties(self) -> Dict[str, bool]:
        """ Dictionary of the UserAccountControl properties. """
        return self.__properties

    @property
    def value(self) -> int:
        """ The intger value of the properties. """
        return sum(
            self.__flag_values[key] for key, val in self.properties.items() if val
        )



import struct

from typing import Any, Optional, Tuple


class SID:
    """
    A class for representing a Security Identifier, that identifies users,
    groups, and computer accounts on a Microsoft Windows platform.

    :param str|None str_rep: a string representation of a SID.
    :param bytes|None bytes_le: a bytes representation of a SID in little-endian
        byte order.
    :raises TypeError: when the type of the parameters are invalid, or both
        parameters are given.
    :raises ValueError: when the given parameter cannot be parsed as a valid
        SID.
    """

    def __init__(
        self, str_rep: Optional[str] = None, bytes_le: Optional[bytes] = None
    ) -> None:
        if str_rep is not None and bytes_le is not None:
            raise TypeError(
                "Only one of the `str_rep` or `bytes_le` parameter must be given"
            )
        if str_rep is not None:
            try:
                if not isinstance(str_rep, str):
                    raise TypeError("The `str_rep` parameter must be a string")
                parts = str_rep.split("-")
                if parts[0] != "S":
                    raise ValueError()
                self.__revision = int(parts[1])
                self.__identifier_authority = (
                    int(parts[2], 16) if "0x" in parts[2] else int(parts[2])
                )
                self.__subauthorities = tuple(int(sub) for sub in parts[3:])
            except (ValueError, IndexError):
                raise ValueError(f"String `{str_rep}` is not a valid SID")
        if bytes_le is not None:
            try:
                if not isinstance(bytes_le, bytes):
                    raise TypeError("The `bytes_le` parameter must be bytes")
                self.__revision, subauth_count, *identifier_auth = struct.unpack(
                    "<BB6B", bytes_le[:8]
                )
                self.__subauthorities = struct.unpack(
                    f"<{subauth_count}I", bytes_le[8 : 8 + (subauth_count * 4)]
                )
                self.__identifier_authority = sum(
                    num << ((5 - i) * 8) for i, num in enumerate(identifier_auth)
                )
            except struct.error as err:
                raise ValueError(f"Not a valid binary SID, {err}")

    def __str__(self) -> str:
        """Return the string format of the SID."""
        ident_auth = (
            hex(self.__identifier_authority)
            if self.__identifier_authority > 2 ** 32
            else self.__identifier_authority
        )
        subauths = (
            "-".join(str(sub) for sub in self.__subauthorities)
            if self.__subauthorities
            else "0"
        )
        return f"S-1-{ident_auth}-{subauths}"

    def __repr__(self) -> str:
        """The representation of SID class."""
        return f"<{self.__class__.__name__}: {str(self)}>"

    def __eq__(self, other: Any) -> bool:
        """
        Check equality of two SIDs by their identifier_authority and list
        of subauthorities, or if the other object is a string than by their
        string formats.
        """
        if isinstance(other, SID):
            return (
                self.revision == other.revision
                and self.identifier_authority == other.identifier_authority
                and self.subauthorities == other.subauthorities
            )
        elif isinstance(other, str):
            return str(self) == other
        else:
            return NotImplemented

    def __hash__(self) -> int:
        return hash(str(self))

    @property
    def revision(self) -> int:
        """The revision level of the SID."""
        return self.__revision

    @property
    def identifier_authority(self) -> int:
        """
        The indentifier that indicates the authority under which
        the SID was created.
        """
        return self.__identifier_authority

    @property
    def subauthorities(self) -> Tuple[int, ...]:
        """
        A tuple of subauthorities that uniquely identifies a principal
        relative to the identifier authority.
        """
        return self.__subauthorities

    @property
    def bytes_le(self) -> bytes:
        """The byte format of the SID in little-endian byte order."""
        subauth_count = len(self.subauthorities)
        identifier_auth = [
            item for item in struct.pack(">Q", self.identifier_authority)[2:]
        ]
        return struct.pack(
            f"<BB6B{subauth_count}I",
            self.revision,
            subauth_count,
            *identifier_auth,
            *self.subauthorities,
        )

    @property
    def sddl_alias(self) -> Optional[str]:
        """
        The string SDDL alias of the SID if it exists, otherwise it's None.
        """
        aliases = {
            "S-1-1-0": "WD",
            "S-1-15-2-1": "AC",
            "S-1-16-12288": "HI",
            "S-1-16-16384": "SI",
            "S-1-16-4096": "LW",
            "S-1-16-8192": "ME",
            "S-1-16-8448": "MP",
            "S-1-3-0": "CO",
            "S-1-3-1": "CG",
            "S-1-3-4": "OW",
            "S-1-5-10": "PS",
            "S-1-5-11": "AU",
            "S-1-5-12": "RC",
            "S-1-5-14": "IU",
            "S-1-5-18": "SY",
            "S-1-5-19": "LS",
            "S-1-5-2": "NU",
            "S-1-5-20": "NS",
            "S-1-5-32-544": "BA",
            "S-1-5-32-545": "BU",
            "S-1-5-32-546": "BG",
            "S-1-5-32-547": "PU",
            "S-1-5-32-548": "AO",
            "S-1-5-32-549": "SO",
            "S-1-5-32-550": "PO",
            "S-1-5-32-551": "BO",
            "S-1-5-32-552": "RE",
            "S-1-5-32-554": "RU",
            "S-1-5-32-555": "RD",
            "S-1-5-32-556": "NO",
            "S-1-5-32-558": "MU",
            "S-1-5-32-559": "LU",
            "S-1-5-32-568": "IS",
            "S-1-5-32-569": "CY",
            "S-1-5-32-573": "ER",
            "S-1-5-32-574": "CD",
            "S-1-5-32-575": "RA",
            "S-1-5-32-576": "ES",
            "S-1-5-32-577": "MS",
            "S-1-5-32-578": "HA",
            "S-1-5-32-579": "AA",
            "S-1-5-32-580": "RM",
            "S-1-5-33": "WR",
            "S-1-5-6": "SU",
            "S-1-5-7": "AN",
            "S-1-5-84-0-0-0-0-0": "UD",
            "S-1-5-9": "ED",
        }
        domain_aliases = {
            498: "RO",
            500: "LA",
            501: "LG",
            512: "DA",
            513: "DU",
            514: "DG",
            515: "DC",
            516: "DD",
            517: "CA",
            518: "SA",
            519: "EA",
            520: "PA",
            522: "CN",
        }
        alias = aliases.get(str(self), None)
        if alias:
            return alias
        elif self.identifier_authority == 5 and self.subauthorities[0] == 21:
            alias = domain_aliases.get(self.subauthorities[-1], None)
        return alias

    @property
    def size(self):
        """The binary size of the SID in bytes."""
        return 8 + len(self.subauthorities) * 4



import os
import uuid

import pytest

from bonsai.active_directory import ACL
from bonsai.active_directory.acl import ACE, ACEFlag, ACEType, ACERight, ACLRevision


def test_ace_from_binary():
    """ Test ACE's from_binary method. """
    input_data = b"\x05\nH\x00\x07\x00\x00\x00\x03\x00\x00\x00\x01\xc9u\xc9\xealoK\x83\x19\xd6\x7fED\x95\x06\x14\xcc(H7\x14\xbcE\x9b\x07\xado\x01^_(\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\x15\xb3LK\xd2\xfb\x90s\xc2\xdf9\xb9\\\x04\x00\x00"
    with pytest.raises(TypeError):
        _ = ACE.from_binary(0)
    with pytest.raises(TypeError):
        _ = ACE.from_binary("INVALID")
    with pytest.raises(ValueError):
        _ = ACE.from_binary(b"\x05\nH\x00\x07\x00\x00\x00")
    ace = ACE.from_binary(input_data)
    assert ace.type == ACEType.ACCESS_ALLOWED_OBJECT
    assert ace.mask == 7
    assert ace.rights == {
        ACERight.DS_CREATE_CHILD,
        ACERight.DS_DELETE_CHILD,
        ACERight.ACTRL_DS_LIST,
    }
    assert ace.size == len(input_data)
    assert ace.flags == {ACEFlag.INHERIT_ONLY, ACEFlag.CONTAINER_INHERIT}
    assert sum(ace.flags) == 10
    assert ace.trustee_sid == "S-1-5-21-1263317781-1938881490-3107577794-1116"
    assert ace.object_type == uuid.UUID("c975c901-6cea-4b6f-8319-d67f45449506")
    assert ace.inherited_object_type == uuid.UUID(
        "4828cc14-1437-45bc-9b07-ad6f015e5f28"
    )
    assert ace.application_data == b""


def test_ace_to_binary():
    """ Test ACE's from_binary method. """
    input_data = b"\x05\nH\x00\x07\x00\x00\x00\x03\x00\x00\x00\x01\xc9u\xc9\xealoK\x83\x19\xd6\x7fED\x95\x06\x14\xcc(H7\x14\xbcE\x9b\x07\xado\x01^_(\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\x15\xb3LK\xd2\xfb\x90s\xc2\xdf9\xb9\\\x04\x00\x00"
    assert ACE.from_binary(input_data).to_binary() == input_data

def test_acl_from_binary():
    """ Test ACL's from_binary method. """
    with pytest.raises(TypeError):
        _ = ACL.from_binary(True)
    with pytest.raises(TypeError):
        _ = ACL.from_binary("INVALID")
    with pytest.raises(ValueError):
        _ = ACL.from_binary(
            b"\x05\nH\x00\x07\x00\x00\x00\x03\x00\x00\x00\x01\xc9u\xc9\xealoK\x83\x19\xd6\x7fED\x95\x06\x14\xcc(H7\x14\xbcE\x9b\x07\xado\x01^_("
        )
    curdir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(curdir, "testenv", "acl-sample.bin"), "rb") as data:
        input_data = data.read()
        acl = ACL.from_binary(input_data)
        assert acl.size == len(input_data)
        assert acl.revision == ACLRevision.ACL_REVISION_DS
        assert len(acl.aces) == 85
        assert all(isinstance(ace, ACE) for ace in acl.aces)
        assert acl.aces[0].type == ACEType.ACCESS_ALLOWED_OBJECT
        assert acl.aces[0].inherited_object_type == uuid.UUID(
            "4828cc14-1437-45bc-9b07-ad6f015e5f28"
        )
        assert acl.aces[-1].type == ACEType.ACCESS_ALLOWED
        assert (
            acl.aces[0].trustee_sid == "S-1-5-21-1263317781-1938881490-3107577794-1116"
        )
        assert acl.aces[-1].trustee_sid == "S-1-5-18"


def test_acl_to_binary():
    """ Test ACL's from_binary method. """
    curdir = os.path.abspath(os.path.dirname(__file__))
    with open(os.path.join(curdir, "testenv", "acl-sample.bin"), "rb") as data:
        input_data = data.read()
        assert ACL.from_binary(input_data).to_binary() == input_data

def test_str():
    """ Test __str__ method. """
    input_data = b"\x05\nH\x00\x07\x00\x00\x00\x03\x00\x00\x00\x01\xc9u\xc9\xealoK\x83\x19\xd6\x7fED\x95\x06\x14\xcc(H7\x14\xbcE\x9b\x07\xado\x01^_(\x01\x05\x00\x00\x00\x00\x00\x05\x15\x00\x00\x00\x15\xb3LK\xd2\xfb\x90s\xc2\xdf9\xb9\\\x04\x00\x00"
    ace = ACE.from_binary(input_data)
    assert (
        str(ace)
        == "(OA;CIIO;CCDCLC;c975c901-6cea-4b6f-8319-d67f45449506;4828cc14-1437-45bc-9b07-ad6f015e5f28;S-1-5-21-1263317781-1938881490-3107577794-1116)"
    )



import sys
from functools import wraps, partial

import pytest
from conftest import get_config, network_delay

from bonsai import LDAPEntry, LDAPClient
import bonsai.errors

try:
    import trio
    from bonsai.trio import TrioLDAPConnection
except ImportError:
    pass
trio = pytest.importorskip("trio")


def trio_test(func):
    @wraps(func)
    def wrapper(*args, **kwargs):
        trio.run(partial(func, *args, **kwargs))

    return wrapper


@pytest.fixture(scope="module")
def tclient():
    """Get an LDAPClient with TrioLDAPConnection async class."""
    cfg = get_config()
    url = "ldap://%s:%s/%s?%s?%s" % (
        cfg["SERVER"]["hostip"],
        cfg["SERVER"]["port"],
        cfg["SERVER"]["basedn"],
        cfg["SERVER"]["search_attr"],
        cfg["SERVER"]["search_scope"],
    )
    cli = LDAPClient(url)
    cli.set_credentials(
        "SIMPLE", user=cfg["SIMPLEAUTH"]["user"], password=cfg["SIMPLEAUTH"]["password"]
    )
    cli.set_async_connection_class(TrioLDAPConnection)
    return cli


@trio_test
async def test_connection(tclient):
    """Test opening a connection."""
    conn = await tclient.connect(True)
    assert conn is not None
    assert conn.closed == False


@trio_test
async def test_search(tclient):
    """Test search."""
    async with tclient.connect(True) as conn:
        res = await conn.search()
        assert res is not None


@trio_test
async def test_add_and_delete(tclient, basedn):
    """Test adding and deleting an LDAP entry."""
    async with tclient.connect(True) as conn:
        entry = LDAPEntry("cn=async_test,%s" % basedn)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
        ]
        entry["sn"] = "async_test"
        try:
            await conn.add(entry)
        except bonsai.errors.AlreadyExists:
            await conn.delete(entry.dn)
            await conn.add(entry)
        except:
            pytest.fail("Unexpected error.")
        res = await conn.search()
        assert entry in res
        await entry.delete()
        res = await conn.search()
        assert entry not in res


@trio_test
async def test_recursive_delete(tclient, basedn):
    """Test removing a subtree recursively."""
    org1 = bonsai.LDAPEntry("ou=testusers,%s" % basedn)
    org1.update({"objectclass": ["organizationalUnit", "top"], "ou": "testusers"})
    org2 = bonsai.LDAPEntry("ou=tops,ou=testusers,%s" % basedn)
    org2.update({"objectclass": ["organizationalUnit", "top"], "ou": "tops"})
    entry = bonsai.LDAPEntry("cn=tester,ou=tops,ou=testusers,%s" % basedn)
    entry.update(
        {"objectclass": ["top", "inetorgperson"], "cn": "tester", "sn": "example"}
    )
    try:
        with (await tclient.connect(True)) as conn:
            await conn.add(org1)
            await conn.add(org2)
            await conn.add(entry)
            with pytest.raises(bonsai.errors.NotAllowedOnNonleaf):
                await conn.delete(org1.dn)
            await conn.delete(org1.dn, recursive=True)
            res = await conn.search(org1.dn, 2)
            assert res == []
    except bonsai.LDAPError as err:
        pytest.fail("Recursive delete is failed: %s" % err)


@trio_test
async def test_modify_and_rename(tclient, basedn):
    """Test modifying and renaming LDAP entry."""
    async with tclient.connect(True) as conn:
        entry = LDAPEntry("cn=async_test,%s" % basedn)
        entry["objectclass"] = [
            "top",
            "inetOrgPerson",
            "person",
            "organizationalPerson",
        ]
        entry["sn"] = "async_test"
        oldname = "cn=async_test,%s" % basedn
        newname = "cn=async_test2,%s" % basedn
        res = await conn.search(newname, 0)
        if res:
            await res[0].delete()
        try:
            await conn.add(entry)
        except bonsai.errors.AlreadyExists:
            await conn.delete(entry.dn)
            await conn.add(entry)
        except:
            pytest.fail("Unexpected error.")
        entry["sn"] = "async_test2"
        await entry.modify()
        await entry.rename(newname)
        res = await conn.search(entry.dn, 0, attrlist=["sn"])
        assert entry["sn"] == res[0]["sn"]
        res = await conn.search(oldname, 0)
        assert res == []
        await conn.delete(entry.dn)


@trio_test
async def test_obj_err(tclient, basedn):
    """Test object class violation error."""
    entry = LDAPEntry("cn=async_test,%s" % basedn)
    entry["cn"] = ["async_test"]
    with pytest.raises(bonsai.errors.ObjectClassViolation):
        async with tclient.connect(True) as conn:
            await conn.add(entry)


@trio_test
async def test_whoami(tclient):
    """Test whoami."""
    async with tclient.connect(True) as conn:
        cfg = get_config()
        obj = await conn.whoami()
        expected_res = [
            "dn:%s" % cfg["SIMPLEAUTH"]["user"],
            cfg["SIMPLEAUTH"]["adusername"],
        ]
        assert obj in expected_res


@pytest.mark.timeout(18)
@trio_test
async def test_connection_timeout(tclient):
    """Test connection timeout."""
    with network_delay(6.0):
        with pytest.raises(bonsai.errors.TimeoutError):
            await tclient.connect(True, timeout=8.0)


@pytest.mark.timeout(18)
@trio_test
async def test_search_timeout(tclient):
    """Test search timeout."""
    async with tclient.connect(True) as conn:
        with network_delay(5.1):
            with pytest.raises(bonsai.errors.TimeoutError):
                await conn.search(timeout=4.0)


@trio_test
async def test_paged_search(tclient, basedn):
    """Test paged results control."""
    tclient.auto_page_acquire = False
    search_dn = "ou=nerdherd,%s" % basedn
    async with tclient.connect(True) as conn:
        res = await conn.paged_search(search_dn, 1, page_size=2)
        for ent in res:
            assert isinstance(ent, bonsai.LDAPEntry)
        page = 1  # First page is already acquired.
        while True:
            if len(res) > 2:
                pytest.fail("The size of the page is greater than expected.")
            msgid = res.acquire_next_page()
            if msgid is None:
                break
            res = await conn.get_result(msgid)
            page += 1
        assert page == 3


@trio_test
async def test_paged_search_with_auto_acq(tclient, basedn):
    """Test paged search with auto page acquiring."""
    search_dn = "ou=nerdherd,%s" % basedn
    async with tclient.connect(True) as conn:
        cnt = 0
        result = await conn.paged_search(search_dn, 1, page_size=3)
        async for item in result:
            assert isinstance(item, LDAPEntry)
            cnt += 1
        assert cnt == 6


@trio_test
async def test_async_with(tclient):
    """Test async with context manager (with backward compatibility)."""
    mgr = tclient.connect(True)
    aexit = type(mgr).__aexit__
    aenter = type(mgr).__aenter__(mgr)

    conn = await aenter
    try:
        assert conn.closed == False
        _ = await conn.whoami()
    except:
        if not (await aexit(mgr, *sys.exc_info())):
            raise
    else:
        await aexit(mgr, None, None, None)
    assert conn.closed
